#!/usr/bin/perl

use strict;
use warnings;
use Getopt::Std;
use Pod::Usage;

use Tk;
use Tk::Photo;
use Tk::PNG;
use Tk::DialogBox;
use Tk::HList;

use Math::Trig qw/pi tan atan/;
use POSIX qw/ceil floor/; # only for lnglat <-> coords

use threads;
use threads::shared;
use Thread::Queue;
use utf8; # for geocoding

use Text::CSV_XS; # for poi

use DownImgs;

my $VERSION = '1.3.2';

my %opts;
getopts( 'h', \%opts );

#pod2usage( -verbose => 2, -noperldoc => 1 ) if $opts{h};

=head1 NAME

pgmap

=head1 SYNOPSIS

pgmap.pl


=cut

#  Todo: regional selection .. OK
#     fix getting corrupted after download and update ...working OK?
#     fix getting corrupted after window resize .. tentatively OK
#     reorganize thread (to decrese number of thread) ... seems OK
#     widgets (zoom, mag, cursor on/off, grid on/off, download).. tentatively OK
#     menu (do not support?)
#     map functions (jump, show lat/lng, copy-buffer, remove map).. partially OK
#     initial download (in conjuction with menu? or quick start button?) just implemented
#     status message while loading ... tentatively OK
#     support cygwin perl as much as possible ... seems working
#     location memory (up to 10?)
#     POI support (filtering, list window) ... priliminary OK
#     select all ... OK
#     GPS support ... preliminary implemetation done (GUI widgets needed .. OK)
#     show x,y over image file (or only show property with right-click menu)
#     handling cursor key (currently get corrupted) ... OK
#     support lnglat in deg.min.sec notation ... OK
#     geocoding support .. OK
#     runtime configuration of image_backup_level
#     optimize show-backup-mode (to reduce the number of generated images)
#
#     canvas layer management (display order) .... looks OK
#     search for backup file in two-level? (not needed?) .. tentatively OK
#     debuging information, logging
#     Pod, version
#
#     File handling menu (show location, remove?)
#     Error handlings
#       address lookup
#       download error ... need to skip saving images in error case
#       confuration support to skip fetchver


# disable buffering in STDOUT (for debugging)
select( ( select( STDOUT ), $| = 1 )[$[] );

# configurations

my $basedir = 'imgs';
my $icondir = 'pgmap_data';

my $img_size = 256;
my $cursor_size = 16;
my @zoom_range = ( 1, 20 );
my @magnify_range = ( -5, 5 );

my $cursor_mode = 1;
my $border_mode = 1;
my $image_backup_level = 1;

my $gps_connect = 0;
share( $gps_connect );

my $gps_track_mode = 1;
my $gps_initialized = 0;
my $gps_valid = 0;
my $gps_status;
my $gps_distance = '';
my $gps_blink = 0;
my $gps_device;
my $gps_baudrate;
my ( @gps_map_coords, $gps_lnglat );
my $gps_control;

my $timer_interval = 100; # in ms
my $max_download_thread = 4;

my $save_status_mode = 1;  # auto save mode
my $running = 1;

my $cursorkey_scroll = 20;
my $shift_cursorkey_scroll = 200;

my $initial_lnglat = "139.770298, 35.689021";
my $initial_zoom = 13;
my $initial_mag = 0;

load_config();

## initial setting

my $zoom = Log2->new( $initial_zoom );
my $magnify = Log2->new( $initial_mag );
my @map_coords = lnglat2coords( split( /,/, $initial_lnglat ) );

# work variables

my @canvas_size = ( 1, 1 ); # initial size before being shown
my @canvas_center;
my ( %map, %map_id );
my ( %blank_image );
my ( @download_task );
my ( @mouse, @select_region, $select_id, @select_offset );
my ( $lnglat, $lnglat_entry );
my ( @poi, %poi_id, $poilist_win, $poidata_win, %poi_icon, $poidata_hlist, $poi_index, $poi_select );
my ( @prev_status );
my ( @poi_target ) = ( 0, 0 );

my $download_queue = Thread::Queue->new;
my $complete_queue = Thread::Queue->new;
my $address_lookup_queue = Thread::Queue->new;
my $address_lookup_complete_queue = Thread::Queue->new;
my $gps_queue = Thread::Queue->new;
my $downimgs = DownImgs->new;
my $status_message = '';
my $address_lookup_initialized = 0;

# count number of image objects
my $debug_images;

load_prev_status();

load_poi();

# TK init

my $top = MainWindow->new;
$top->bind( "<Configure>", \&change_size );
$top->geometry( "800x600" );
$top->protocol( 'WM_DELETE_WINDOW', \&exit_program );

set_widgets();

my $canvas = $top->Canvas( -xscrollincrement => 1, -yscrollincrement => 1 )
  ->pack( -expand => 1, -fill => 'both', -side => 'top' );

set_global_bindings();

# cursor and selection rectangle in reverse order
$canvas->createLine( 0, 0, 0, 0, -tags => 'partition_data' ); # data should be lower than this
$canvas->createLine( 0, 0, 0, 0, -tags => 'partition_poi' ); # poi should be lower than this
$canvas->createLine( 0, 0, 0, 0, -tags => 'cursor_v' );
$canvas->createLine( 0, 0, 0, 0, -tags => 'cursor_h' );
$canvas->createRectangle( 0, 0, 0, 0, -tags => 'select' );
$canvas->createLine( 0, 0, 0, 0, -fill => 'blue', -width => 2, -tags => 'gpstarget' );
$canvas->createLine( 0, 0, 0, 0, -fill => 'purple', -width => 2, -tags => 'poitarget' );
$canvas->createOval( 0, 0, 0, 0, -tags => 'gps' );

# start periodical task
taskmanager();

# need to be modified so that gps_thread is executed regardless of $gps_track_mode
gps_connect() if $gps_connect;
#

MainLoop;

# -----------------------------------------------------------------
#  widgets and bindings
# -----------------------------------------------------------------

sub set_widgets {
  my $control_base = $top->Frame->pack( -fill => 'x', -side => 'top' );

  my $control = $control_base->Frame->pack( -fill => 'x', -side => 'top' );
  $gps_control = $control_base->Frame;

  $control->Label( -text => 'Zoom:' )->pack( -side => 'left' );
  my $zoomMenu = $control->Optionmenu( -variable => \$zoom->{exp},
				       -command => \&zoom_update )->pack( -side => 'left' );
  $zoomMenu->addOptions( $zoom_range[0] .. $zoom_range[1] );
  $control->Button( -text => '+', -width => 1, -command => \&zoom_in )->pack( -side => 'left', -padx => 4 );
  $control->Button( -text => '-', -width => 1, -command => \&zoom_out )->pack( -side => 'left', -padx => 4 );

  $control->Label( -text => 'Mag:' )->pack( -side => 'left' );
  my $magMenu = $control->Optionmenu( -variable => \$magnify->{exp},
				      -command => \&mag_update )->pack( -side => 'left' );
  $magMenu->addOptions( $magnify_range[0] .. $magnify_range[1] );
  $control->Button( -text => '+', -width => 1, -command => \&mag_in )->pack( -side => 'left', -padx => 4 );
  $control->Button( -text => '-', -width => 1, -command => \&mag_out )->pack( -side => 'left', -padx => 4 );
  $control->Checkbutton( -text => 'Grid', -variable => \$border_mode,
			 -command => \&update_border )->pack( -side => 'left', -padx => 4 );
  $control->Checkbutton( -text => 'Cursor', -variable => \$cursor_mode,
			 -command => \&update_cursor )->pack( -side => 'left', -padx => 4 );

  $lnglat_entry = $control->Entry( -width => 20, -textvariable => \$lnglat )
    ->pack( -side => 'left', -padx => 4 );
  $lnglat_entry->bind( "<Return>", \&move_lnglat );

  $control->Button( -text => 'Download Map', -command => \&download_maps )
    ->pack( -side => 'left', -padx => 4 );

  # just for debug
#  $control->Label( -textvariable => \$debug_images )->pack( -side => 'left' );

  $control->Label( -textvariable => \$status_message, -anchor => 'w' )->pack( -side => 'left', -fill => 'x', -expand => 1 );

  $control->Button( -text => '?', -command => \&show_help )->pack( -side => 'left', -padx => 4 );

  $gps_control->Checkbutton( -text => 'Track', -variable => \$gps_track_mode, 
			     -command => \&update_scroll )->pack( -side => 'left' );

  $gps_control->Label( -text => 'Status' )->pack( -side => 'left' );
  $gps_control->Label( -width => 8, -textvariable => \$gps_status, -relief => 'sunken', -anchor => 'w' )
    ->pack( -side => 'left', -padx => 4 );
  $gps_control->Label( -width => 20, -textvariable => \$gps_lnglat, -relief => 'sunken', -anchor => 'w' )
    ->pack( -side => 'left', -padx => 4 );
  $gps_control->Button( -text => 'Disconnect' )->pack( -side => 'left' );

  $gps_control->pack( -fill => 'x', -side => 'top' ); # need to pack here at the begining

  unless ( $gps_connect ) {
    $gps_control->packForget;
  }
}

sub set_global_bindings {
  my %bindings = (
		  'KeyPress-i' => \&zoom_in,
		  'KeyPress-o' => \&zoom_out,
		  'KeyPress-I' => \&mag_in,
		  'KeyPress-O' => \&mag_out,
		  'KeyPress-period' => \&toggle_cursor,
		  'KeyPress-exclam' => \&toggle_border,
		  'KeyPress-D' => \&download_maps,
		  'KeyPress-A' => \&select_all,
		  'KeyPress-j' => sub { $lnglat_entry->focus },
		  'KeyPress-Q' => \&exit_program,
		  'KeyPress-R' => \&apply_prev_status,
		  'KeyPress-g' => \&gps_connect,
		  'KeyPress-G' => \&gps_disconnect,
		  'KeyPress-t' => sub { $gps_track_mode = !$gps_track_mode; update_scroll() },
		  'KeyPress-h' => \&show_help,
		  'KeyPress-question' => \&show_help,
		  'KeyPress-a' => \&address_entry,
		  'KeyPress-Up' => [\&move_map, 0, $cursorkey_scroll],
		  'KeyPress-Down' => [\&move_map, 0, -$cursorkey_scroll],
		  'KeyPress-Right' => [\&move_map, -$cursorkey_scroll, 0],
		  'KeyPress-Left' => [\&move_map, $cursorkey_scroll, 0],
		  'Shift-KeyPress-Up' => [\&move_map, 0, $shift_cursorkey_scroll],
		  'Shift-KeyPress-Down' => [\&move_map, 0, -$shift_cursorkey_scroll],
		  'Shift-KeyPress-Right' => [\&move_map, -$shift_cursorkey_scroll, 0],
		  'Shift-KeyPress-Left' => [\&move_map, $shift_cursorkey_scroll, 0],
		  'KeyPress-p' => \&show_poilist_win,
		  'KeyPress-P' => \&show_poidata_win,
		  'KeyPress-m' => \&set_mark,
		  '1', [\&mouse_press, Ev('x'), Ev('y')],
		  'B1-Motion', [\&mouse_motion, Ev('x'), Ev('y')],
		  'B1-ButtonRelease', [\&mouse_release, Ev('x'), Ev('y')],
		  'Double-1', [\&mouse_double_click, Ev('x'), Ev('y')],
		  'Shift-Double-1', [\&mouse_shift_double_click, Ev('x'), Ev('y')],
		  'Control-KeyPress-Q' => \&quickstart_download,
		  'KeyPress-d' => \&debug_func,
		 );

  # disable default cursor key bindings, then define bindings
  foreach my $key ( 'Up', 'Down', 'Right', 'Left' ) {
    $top->bind( 'Tk::Canvas', "<$key>" => '' );
  }

  foreach my $sequence ( keys %bindings ) {
    $canvas->CanvasBind( "<$sequence>", $bindings{$sequence} );
  }
}

# -----------------------------------------------------------------------
#  handling download and address lookup
# -----------------------------------------------------------------------

sub taskmanager {
  # update image of which download complete
  while ( $complete_queue->pending ) {
    my $arg = $complete_queue->dequeue;
    my ( $z, $x, $y ) = split( /,/, $arg );

    # skip update if zoom of downloaded is different from current one
    next unless $z == $zoom->ex;

    my $src = getImgURL( $z, $x, $y );
    if ( -r $src ) {
      my $image = get_image( $magnify->ex, 0, 0, 0, $src );

      $canvas->itemconfigure( $map{ $x, $y }->{image_id}, -image => $image );
      $canvas->itemconfigure( $map{ $x, $y }->{border_id}, -outline => 'gray' );
      $map{ $x, $y }->{backup} = 0;
      $map{ $x, $y }->{selected} = 0;

      # remove image resource
      if ( $map{ $x, $y }->{blank} ) {
	# only register
	$map{ $x, $y }->{blank} = 0;
      } else {
	# discard backup image
	if ( defined $map{ $x, $y }->{image} && defined !$map{ $x, $y }->{blank} ) {
	  $map{ $x, $y }->{image}->delete;
	  $debug_images--;
	}
      }
      $map{ $x, $y }->{image} = $image;
    }
  }

  while ( $address_lookup_complete_queue->pending ) {
    # address lookup
    my $arg = $address_lookup_complete_queue->dequeue;
    if ( defined $arg ) {
      $lnglat = $arg;
      move_lnglat();
    }
    $status_message = '';
  }

  if ( my $pending = $download_queue->pending ) {
    $status_message = "downloading.. ($pending maps left)";
  } else {
    if ( $status_message =~ /^downloading/ ) {
      $status_message = '';
    }
  }

# GPS support
  my $gpsout;
  while ( $gps_queue->pending ) {
    $gpsout = $gps_queue->dequeue;
  }
  if ( defined $gpsout ) {
    my ( $lng, $lat, $quality, $ngps ) = split( /,/, $gpsout );
    $gps_status = "$ngps:" . ( '', 'GPS', 'DGPS' )[$quality];
    if ( $quality ) {
      $gps_valid = 1;

     # udpate @map_coords here
      @gps_map_coords = lnglat2coords( $lng, $lat );
      $gps_lnglat = sprintf( "%.6f %.6f", $lng,$lat );
      if ( $gps_track_mode ) {
	@map_coords = @gps_map_coords;
	draw_map();
	update_scroll();
      } else {
	update_gpspos();
      }
    } else {
      $gps_valid = 0;
    }
  }

  # blink GPS cursor if enabled
  if ( $gps_connect ) {
    if ( $gps_valid ) {
      if ( time % 2 ) {
	if ( $gps_blink == 0 ) {
	  $canvas->itemconfigure( 'gps', -fill => '#FF0000' );
	  $gps_blink = 1;
	}
	#      print "blink 0\n";
      } else {
	if ( $gps_blink == 1 ) {
	  $canvas->itemconfigure( 'gps', -fill => '#800000' );
	  $gps_blink = 0;
	}
	#      print "blink 1\n";
      }
    } else {
      # stop blinking once ...
      if ( $gps_blink == 0 ) {
	$canvas->itemconfigure( 'gps', -fill => '#FF0000' );
	$gps_blink = 1;
      }
    }
  }

  $top->after( $timer_interval, \&taskmanager );
}

# Background threads to download map in queue
sub download_map_thread {
  my ( $thread_index ) = @_;# for debugging

  while ( my $arg = $download_queue->dequeue ) {
    # download
    $downimgs->down_map1( split( /,/, $arg ) );

    # push results in queue
    $complete_queue->enqueue( $arg );
  }
}

sub address_lookup_thread {
  while ( my $arg = $address_lookup_queue->dequeue ) {
    # address lookup
    my $lnglat = $downimgs->geocoding( $arg );

    $address_lookup_complete_queue->enqueue( $lnglat );
 }
}

# Enqueue download tasks
sub download_maps {
  # fetch google map version only once
  init_download() or return;

  foreach my $xy ( sort sortxy keys %map ) {
    if ( $map{$xy}->{selected} ) {
      # only if selected
      my ( $x, $y ) = split( /$;/, $xy );
      $download_queue->enqueue( join( ',', $zoom->ex, $x, $y ) );
    }
  }
}

sub init_address_lookup {
  # only once ...
  return 1 if $address_lookup_initialized;

  threads->new( \&address_lookup_thread, $_ ) or return;
  $address_lookup_initialized = 1;
}


sub init_download {
  # do it only once;
  return 1 if $downimgs->getver;

  $status_message = 'Getting GMap version';
  $canvas->update;
  unless ( $downimgs->fetchver ) {
    $status_message = 'Cannot connect to Gmap';
    return;
  }
  $status_message = '';

  foreach ( 0 .. $max_download_thread - 1 ) {
    # create new thread here
    threads->new( \&download_map_thread, $_ );
  }

  return 1;
}

sub quickstart_download {
  init_download() or return;

  my @init_download_list = (
			    [13,57,22], [13,57,23], [13,57,24], [13,57,25],
			    [13,56,23],[13,56,24], [13,56,25], [13,55,25], 
			    [13,55,26], [13,54,25], [13,54,26],);

  foreach my $map1 ( @init_download_list ) {
    $download_queue->enqueue( join( ',', @$map1 ) );
  }
}

# -----------------------------------------------------------------------
#  draw maps
# -----------------------------------------------------------------------

sub draw_map {
  # ignore if called by initial widget callbacks
  return unless @canvas_center;

  my @map_index = ( int( $map_coords[0] / $img_size / $zoom->p2 ), int( $map_coords[1] / $img_size / $zoom->p2 ) );
  my @image_num = ( int( $canvas_center[0] / ( $img_size * $magnify->p2 ) ) + 1,
		    int( $canvas_center[1] / ( $img_size * $magnify->p2 ) ) + 1 );

  my @map_pt;


  my @coords_range = ( ( $map_index[0] - $image_num[0] ) * $img_size * $zoom->p2,
		       ( $map_index[1] - $image_num[1] ) * $img_size * $zoom->p2,
		       ( $map_index[0] + $image_num[0] + 1 ) * $img_size * $zoom->p2,
		       ( $map_index[1] + $image_num[1] + 1 ) * $img_size * $zoom->p2,
		     );

  foreach my $y ( $map_index[1] - $image_num[1] .. $map_index[1] + $image_num[1] ) {
    $map_pt[1] = $y * $img_size * $magnify->p2;
    foreach my $x ( $map_index[0] - $image_num[0] .. $map_index[0] + $image_num[0] ) {
      next if defined $map{$x, $y}->{image};
      $map_pt[0] = $x * $img_size * $magnify->p2;

      my ( $border_color, $image );
      my ( $backup, $blank, $selected ) = ( 0, 0, $map{$x, $y}->{selected} ? 1 : 0 );
      my $tag = 'data';

      my $src = getImgURL( $zoom->ex, $x, $y );
      if ( -r $src ) {
	$image = get_image( $magnify->ex, 0, 0, 0, $src );
	$border_color = 'gray';
#	print "[$x,$y] image found $map_pt[0], $map_pt[1]\n";
      } else {
	foreach my $level ( 1 .. $image_backup_level ) {
	  $src = getImgURL( $zoom->ex + $level, int( $x / 2 ** $level ), int( $y / 2 ** $level ) );
	  if ( -r $src ) {
#	    print "[$x,$y] backup $level found $map_pt[0], $map_pt[1]\n";
	    $image = get_image( $magnify->ex, $level, $x % 2 ** $level, $y % 2 ** $level, $src );
	    $tag .= ' backup';
	    $backup = 1;
	    last;
	  }
	}
	unless ( defined $image ) {
#	  print "[$x,$y]blank $map_pt[0], $map_pt[1]\n";
	  $image = get_image( $magnify->ex, 0, 0, 0, undef );
	  $blank = 1;
	  $tag .= ' blank';
	}
	$border_color = $selected ? 'red' : 'blue';
      }

      my $image_id = $canvas->createImage( @map_pt, -image => $image, -anchor => 'nw', -tags => $tag );
      $canvas->bind( $image_id, "<Shift-Button-1>", [\&mouse_click_data, Ev('x'), Ev('y')] );
      $canvas->bind( $image_id, "<Shift-ButtonRelease-1>", [\&mouse_release_data, Ev('x'), Ev('y')] );
      $canvas->bind( $image_id, "<Control-Button-1>", [\&mouse_click_data, Ev('x'), Ev('y'), -force_select => 1] );
      $canvas->bind( $image_id, "<Control-ButtonRelease-1>", [\&mouse_release_data, Ev('x'), Ev('y'), -force_select => 1] );
      my $border_id = $canvas->createRectangle( @map_pt,
						$map_pt[0] + $img_size * $magnify->p2 - 1,
						$map_pt[1] + $img_size * $magnify->p2 - 1,
						-outline => $border_color, -tags => 'border' );
      $map_id{$image_id} = $map{$x,$y} = {
					  selected => $selected,
					  backup => $backup,
					  blank => $blank,
					  image_id => $image_id,
					  border_id => $border_id,
					  image => $image,
					  x => $x,
					  y => $y,
					 };
    }
  }

  # int( map_coors / zoom / img_size ) -> map_index
  # map_index * img_size * mag -> pt
  # int( coords / zoom * mag ) -> pt

  foreach my $poi ( @poi ) {
    next if $poi->{zoom} < $zoom->ex;
    next unless $poi->{show};

    foreach my $data ( @{$poi->{data}} ) {
      # id exists and out of range ... delete
      # id not exists and in range ... create
      # otherwise ... do nothing

      my $in_range = $data->{coords}->[0] >= $coords_range[0]
	&& $data->{coords}->[0] <= $coords_range[2]
	  && $data->{coords}->[1] >= $coords_range[1]
	    && $data->{coords}->[1] <= $coords_range[3];

#      my $dcid = defined $data->{cid};
#      print "poilist_id: $poi->{id} in_range $in_range, cid $dcid\n";
      if ( defined $data->{cid} && !$in_range ) {
	# delete
	$canvas->delete( $data->{cid} );
	delete( $poi_id{$data->{cid}} );
	$data->{cid} = undef;
      } elsif ( !defined $data->{cid} && $in_range ) {
	# create
	my @pt = ( $data->{coords}->[0] / $zoom->p2 * $magnify->p2, $data->{coords}->[1] / $zoom->p2 * $magnify->p2 );

	$data->{cid} = $canvas->createImage( $pt[0] + 0, $pt[1] + 0, -image => get_icon( $poi->{icon} ),
					    -anchor => 's', -tags => 'poi' );
	$canvas->bind( $data->{cid}, "<Button-1>", [\&mouse_click_poidata, $data->{cid}, $data] );
	$poi_id{$data->{cid}} = $data;
      }
    }
  }

  $canvas->lower( 'data', 'partition_data' );

  update_scroll();
  update_gpspos();

  update_cursor();
  update_border();
  update_lnglat();

  update_poidata();

  $canvas->Tk::focus;
}

sub update_scroll {
  my @offset = @_;
  @offset = ( 0, 0 ) unless @offset;
  my @map_coords_zoom = ( int( $map_coords[0] / $zoom->p2 ) * $magnify->p2 - $offset[0],
			  int( $map_coords[1] / $zoom->p2 ) * $magnify->p2 - $offset[1] );

  # update cursor location
  $canvas->coords( 'cursor_v', $map_coords_zoom[0], $map_coords_zoom[1] - $cursor_size,
		   $map_coords_zoom[0], $map_coords_zoom[1] + $cursor_size );
  $canvas->coords( 'cursor_h', $map_coords_zoom[0] - $cursor_size, $map_coords_zoom[1],
		   $map_coords_zoom[0] + $cursor_size, $map_coords_zoom[1] );

  # update poi-target if any
  if ( $poi_target[0] > 0 || $poi_target[1] > 0 ) {
    my @target_zoom = ( int( $poi_target[0] / $zoom->p2 ) * $magnify->p2,
			int( $poi_target[1] / $zoom->p2 ) * $magnify->p2 );

    $canvas->coords( 'poitarget', $map_coords_zoom[0], $map_coords_zoom[1],
		     $target_zoom[0], $target_zoom[1] );
  }

  # update gps-target if required
  if ( $gps_connect && @gps_map_coords ) {
    my @gps_map_coords_zoom = ( int( $gps_map_coords[0] / $zoom->p2 ) * $magnify->p2,
				int( $gps_map_coords[1] / $zoom->p2 ) * $magnify->p2 );
    $canvas->coords( 'gpstarget', $gps_map_coords_zoom[0], $gps_map_coords_zoom[1],
		     $map_coords_zoom[0], $map_coords_zoom[1] );
  }

  # reset to origin
  $canvas->xview( moveto => 0 );
  $canvas->yview( moveto => 0 );

  # abstract scroll
  $canvas->xview( scroll => $map_coords_zoom[0] - $canvas_center[0], 'units' );
  $canvas->yview( scroll => $map_coords_zoom[1] - $canvas_center[1], 'units' );
}

sub update_gpspos {
  return unless $gps_connect && @gps_map_coords;

  my @offset = @_;
  @offset = ( 0, 0 ) unless @offset;

  my @gps_map_coords_zoom = ( int( $gps_map_coords[0] / $zoom->p2 ) * $magnify->p2 - $offset[0],
			      int( $gps_map_coords[1] / $zoom->p2 ) * $magnify->p2 - $offset[1] );
  my @map_coords_zoom = ( int( $map_coords[0] / $zoom->p2 ) * $magnify->p2 - $offset[0],
			      int( $map_coords[1] / $zoom->p2 ) * $magnify->p2 - $offset[1] );
  $canvas->coords( 'gps', $gps_map_coords_zoom[0] - 4, $gps_map_coords_zoom[1] - 4,
		   $gps_map_coords_zoom[0] + 4, $gps_map_coords_zoom[1] + 4 );
  $canvas->coords( 'gpstarget', $gps_map_coords_zoom[0], $gps_map_coords_zoom[1],
		   $map_coords_zoom[0], $map_coords_zoom[1] );

}

sub get_icon {
  my ( $icon ) = @_;

  exists $poi_icon{ $icon } ? $poi_icon{ $icon } :
    $poi_icon{ $icon } = $top->Photo( -format => 'png', -file => "$icondir/$icon" );
}

sub get_image {
  my ( $mag, $z, $offset_x, $offset_y, @files ) = @_;

  unless ( defined $files[0] ) {
    # blank image;
    return $blank_image{ $mag } if exists $blank_image{ $mag };

    unless ( exists $blank_image{0} ) {
      # alternative method needed when blank file is missing
      my $blank_file = "$icondir/blank.png";
      if ( -r $blank_file ) {
	$blank_image{0} = $top->Photo( -format => 'png', -file => $blank_file );
      } else {
	# just error message
      }
      $debug_images++;
    }

    if ( $mag > 0 ) {
      $blank_image{$mag} = $top->Photo;
      $blank_image{$mag}->copy( $blank_image{0}, -zoom => 2 ** $mag );
    } elsif ( $mag < 0 ) {
      $blank_image{$mag} = $top->Photo;
      $blank_image{$mag}->copy( $blank_image{0}, -subsample => 2 ** (-$mag) );
    }
    return $blank_image{ $mag };
  }

  my $image = $top->Photo( -format => 'png', -file => $files[0] );
  $debug_images++;

  return $image if $mag == 0 && $z == 0;

=comment

  # case 1: mag = 0, zoom 1 (2x zoom)
    @offset = ( $x % 2 ) * $imgSize / 2, ( $y % 2 ) * $imgSize / 2 );
    $image2->copy( $image, -zoom => ( 2, 2 ), -from => 
	 $offset[0], $offset[1], $offset[0] + $imgSize / 2, $offset[1] + $imgSize / 2 );

  # case 2: mag = -1, zoom 1 (2x zoom, 1/2 mag)
    @offset = ( $x % 2 ) * $imgSize / 2, ( $y % 2 ) * $imgSize / 2 );
    $image2->copy( $image, -zoom => ( 1, 1 ), -from => 
	 $offset[0], $offset[1], $offset[0] + $imgSize / 2, $offset[1] + $imgSize / 2 );

  # case 3: mag = -1, zoom 0 (1/2 mag)
    $image2->copy( $image, -subsample => ( 2, 2 ) )

  # case 4: mag = 0, zoom -1 (1/2 zoom)
    $image2->copy( $image10, -subsample => ( 2, 2 ) )
    $image2->copy( $image11, -subsample => ( 2, 2 ) )
    $image2->copy( $image12, -subsample => ( 2, 2 ) )
    $image2->copy( $image13, -subsample => ( 2, 2 ) )

=cut

  my $zr = $img_size / 2 ** $z;
  my $image2 = $top->Photo;
  $debug_images++;
  if ( $mag + $z >= 0 ) {
    my $r = 2 ** ( $mag + $z );
    $image2->copy( $image, -zoom => ( $r, $r ), -from =>
		    $offset_x * $zr, $offset_y * $zr, 
		    ( $offset_x + 1 ) * $zr, ( $offset_y + 1 ) * $zr );

  } else {
    my $r = 2 ** (-( $mag + $z ));
    $image2->copy( $image, -subsample => ( $r, $r ), -from =>
		    $offset_x * $zr, $offset_y * $zr, 
		    ( $offset_x + 1 ) * $zr, ( $offset_y + 1 ) * $zr );
  }

#  $image2->redither;
  $image->delete;
  $debug_images--;
  $image2;
}

# -----------------------------------------------------------------------
#  GUI handling
# -----------------------------------------------------------------------

sub change_size {
  my @new_canvas_size = ( $canvas->width, $canvas->height );

  if ( $new_canvas_size[0] != $canvas_size[0] || $new_canvas_size[1] != $canvas_size[1] ) {
    @canvas_size = @new_canvas_size;
    @canvas_center = ( int( $canvas_size[0] / 2 ), int( $canvas_size[1] / 2 ) );

    # need to be optimized
    clear_map( -keep_select => 1 );

    draw_map();
  }

  save_status();
}

sub mouse_motion {
  if ( @select_region ) {
    $canvas->coords( 'select',
		     $select_region[0] + $select_offset[0],
		     $select_region[1] + $select_offset[1],
		     $_[1] + $select_offset[0], $_[2] + $select_offset[1] );
  } elsif ( @mouse ) {
    # scroll

    my @diff = ( $_[1] - $mouse[0], $_[2] - $mouse[1] );
#d  print "mv $_[1], $_[2] (diff= @diff)\n";

    update_scroll( @diff );
    update_lnglat( $diff[0] * $zoom->p2 / $magnify->p2 ,$diff[1] * $zoom->p2 / $magnify->p2 );

    # turn off gps_track_mode
    $gps_track_mode = 0;
  }
}

sub mouse_shift_double_click {
  # only centering
  # src = clicked point, dest = center

  my @diff = ( $canvas_center[0] - $_[1], $canvas_center[1] - $_[2] );

  # udpate mapCoords
  @map_coords = ( $map_coords[0] - $diff[0] * $zoom->p2 / $magnify->p2,
		  $map_coords[1] - $diff[1] * $zoom->p2 / $magnify->p2 );

  draw_map();

  save_status();
}

sub mouse_double_click {
  # centering, zoom in and then move to cursor point

  return if $zoom->ex <= $zoom_range[0];

  # 1. src = clicked point, dest = center
  my @diff = ( $canvas_center[0] - $_[1], $canvas_center[1] - $_[2] );
  @map_coords = ( $map_coords[0] - $diff[0] * $zoom->p2 / $magnify->p2,
		  $map_coords[1] - $diff[1] * $zoom->p2 / $magnify->p2 );

  $zoom--;

  # 2. src = center dest = clicked point
  @diff = ( $_[1] - $canvas_center[0], $_[2] - $canvas_center[1] );
  @map_coords = ( $map_coords[0] - $diff[0] * $zoom->p2 / $magnify->p2,
		  $map_coords[1] - $diff[1] * $zoom->p2 / $magnify->p2 );

  clear_map();
  draw_map();

  save_status();
}


sub mouse_press {
  @mouse = @_[1,2];
}

sub mouse_release {
  if ( @select_region ) {
    $canvas->coords( 'select', -2, -2, -1, -1 );
    @select_region = ();
    return;
  }

  return unless @mouse;
  my @diff = ( $_[1] - $mouse[0], $_[2] - $mouse[1] );
  return if $diff[0] == 0 && $diff[1] == 0;

  # udpate mapCoords
  @map_coords = ( $map_coords[0] - $diff[0] * $zoom->p2 / $magnify->p2,
		  $map_coords[1] - $diff[1] * $zoom->p2 / $magnify->p2 );

  @mouse = ();

  draw_map();
  save_status();
}

sub mouse_click_data {
  @select_region = @_[1,2];

  ( $select_id ) = $canvas->find( 'withtag', 'current' );

  @select_offset = (
		    int( $map_coords[0] / $zoom->p2 * $magnify->p2 ) - $canvas_center[0],
		    int( $map_coords[1] / $zoom->p2 * $magnify->p2 ) - $canvas_center[1] );
}

sub mouse_click_poidata {
  return unless $poidata_hlist;
  my ( $canvas, $poi_id ) = @_;
#  print "clicked poiid= $poi_id / $data\n";

  $poidata_hlist->anchorClear();
  $poidata_hlist->selectionClear();
  $poidata_hlist->anchorSet( $poi_id );

  poidata_browse( $poi_id );
}

sub mouse_release_data {
  my @select_release = @_[1,2];
  my %arg = @_[3 .. $#_];

  @select_release = ( $select_offset[0] + $select_release[0],
		      $select_offset[1] + $select_release[1] );

  my $release_id;
  my @release_id = $canvas->find( 'overlapping', @select_release, @select_release );

  foreach my $id ( @release_id ) {
    $release_id = $id, last if exists $map_id{$id};
  }
  unless ( $select_id && $release_id ) {
    $canvas->coords( 'select', -2, -2, -1, -1 );
    return;
  }

  my ( $x0, $y0 ) = ( $map_id{$select_id}->{x}, $map_id{$select_id}->{y} );
  my ( $x1, $y1 ) = ( $map_id{$release_id}->{x}, $map_id{$release_id}->{y} );

  ( $x0, $x1 ) = ( $x1, $x0 ) if $x1 < $x0;
  ( $y0, $y1 ) = ( $y1, $y0 ) if $y1 < $y0;

  foreach my $x ( $x0 .. $x1 ) {
    foreach my $y ( $y0 .. $y1 ) {
      if ( $arg{-force_select} || $map{ $x, $y }->{backup} || $map{ $x, $y }->{blank} ) {
	if ( $map{ $x, $y }->{selected} ) {
	  # backup, selected
	  if ( $map{ $x, $y }->{backup} || $map{ $x, $y }->{blank} ) {
	    $canvas->itemconfigure( $map{ $x, $y }->{border_id}, -outline => 'blue' );
	  } else {
	    $canvas->itemconfigure( $map{ $x, $y }->{border_id}, -outline => 'gray' );
	  }
	  $map{ $x, $y }->{selected} = 0;
	} else {
	  # backup, not selected
	  $canvas->itemconfigure( $map{ $x, $y }->{border_id}, -outline => 'red' );
	  $map{ $x, $y }->{selected} = 1;
	}
      }
    }
  }
  $canvas->coords( 'select', -2, -2, -1, -1 );
  save_status();
}

sub exit_program {
  foreach my $thread ( threads->list ) {
    $thread->detach;
  }
  $running = 0;
#  print "saving status ...";
  save_status();
#  print "done\n";
  exit;
}

sub gps_connect {
  print "== gps connect (gps_connect = $gps_connect) ==\n";

  $gps_control->pack( -side => 'top', -fill => 'x' );

  $gps_connect = 1;
  unless ( $gps_initialized ) {
    threads->new( \&gps_thread );
    # check if thread successfully invoked here

    $gps_initialized = 1;
  } else {
    # kick gps_thread
    lock $gps_connect;
    cond_signal( $gps_connect );
  }
}

sub gps_disconnect {
  print "gps_connect set to 0\n";
  $canvas->coords( 'gps', 0, 0, 0, 0 );
  $canvas->coords( 'gpstarget', 0, 0, 0, 0 );
  $gps_control->packForget;
  $gps_connect = 0;
}

sub gps_thread {
  my $gps_error_count = 0;

  # infinite loop ( state transition: gps_connect <-> gps_update )
  while () {
    # wait until $gps_connect gets 1
    unless ( $gps_connect ) {
      print "gps_connect .. wait for signal (press 'g' key to resume)\n";
      lock $gps_connect;
      cond_wait( $gps_connect );
      print "gps_connect .. resumed\n";
    }

    if ( $gps_device =~ /\// ) {
      # linux based style ( /dev/*** )

      if ( defined $gps_baudrate && system( "stty -F $gps_device $gps_baudrate" ) ) {
	print "Failed to execute stty to $gps_device\n";
	$gps_connect = 0;
	next;
      }
      unless ( open SP, '<', $gps_device ) {
	print "Failed to open '$gps_device'\n";
	$gps_connect = 0;
	next;
      }
      print "opened $gps_device\n";
    } elsif ( $gps_device =~ /com\d+$/i ) {
      # Win32 based style (COMnn)
      my $conf_file = 'serialport.conf';

      unless ( eval "require Win32::SerialPort" ) {
	print "Cannot load Win32::SerialPort module\n";
	$gps_connect = 0;
	next;
      }

      my $com1 = new Win32::SerialPort( $gps_device );
      unless ( defined $com1 ) {
	print "Failed to open COM port '$gps_device' (1)\n";
	$gps_connect = 0;
	next;
      }

      print "Available baudrate: ", join( ',', sort { $a <=> $b } $com1->baudrate ), "\n";

      if ( defined $gps_baudrate ) {
	$com1->baudrate( $gps_baudrate );
      } else {
	# set highest baudrate of available ones
	my $baudrate = ( sort { $b <=> $a } $com1->baudrate )[0];
	print "baudrate is automatically set to $baudrate\n";
	$com1->baudrate( $baudrate );
      }

      $com1->parity( 'none' );
      $com1->databits( 8 ); # needed?
      $com1->stopbits( 1 );
      $com1->handshake('rts');
      $com1->buffers( 1024 * 20, 1024 * 4 ); # read and write buffers
      $com1->read_interval( 800 );
      $com1->read_char_time( 400 );
      $com1->error_msg( 0 ); # no error message like 'Framing error'
      $com1->user_msg( 0 ); # no waring message like 'waining for CTS'

      $com1->write_settings;
      $com1->save( $conf_file );
      $com1->close;

      unless ( tie( *SP, 'Win32::SerialPort', $conf_file ) ) {
	print "Failed to open COM port $gps_device (2)\n";
	$gps_connect = 0;
	next;
      }
      print "opened Win32::SerialPort($gps_device)\n";
    } else {
      print "invalid value in gps_device: $gps_device\n";
      $gps_connect = 0;
      next;
    }

    # fetch GPS data periodically
    my @quality = ( 'No-signal', 'GPS', 'D-GPS' );
    while ( $gps_connect ) {
      $_ = <SP>;
      if ( $? || $! eq 'Permission denied' ) {
	print "device read error ($?;$!)\n";
	last;
      } else {
#	print "read: ($?,$!)$_";
      }

      if ( defined $_ ) {
	$gps_error_count = 0;
      } else {
	# sometimes, $_ can be blank with Win32::Serialport .. ignore several errors in a low
	$gps_error_count++;
	last if $gps_error_count >= 10;
	next;
      }

      if ( /^\$GPGGA/ ) {
	my ( $head, $gtime, $lat1, $lat2, $lng1, $lng2, $quality, $ngps ) = split( /,/ );
	print "time $gtime, LNG:$lng1-$lng2, LAT:$lat1-$lat2 ($quality[$quality]:$ngps)\n";
	if ( $lng1 =~ /^[\d\.]+$/ ) {
	  $lng1 = int( $lng1 / 100 ) + ( $lng1 - int( $lng1 / 100 ) * 100 ) / 60;
	}
	if ( $lat1 =~ /^[\d\.]+$/ ) {
	  $lat1 = int( $lat1 / 100 ) + ( $lat1 - int( $lat1 / 100 ) * 100 ) / 60;
	}

	#      print "$lng1,$lat1\n";

	$gps_queue->enqueue( join( ',', $lng1, $lat1, $quality, $ngps ) );
#	for Demo
#	$gps_queue->enqueue( join( ',', "139.770298", "35.689021", 1, "04" ) );
      }
    }
    close SP;
  }
}

# -----------------------------------------------------------------------
#  callback functions for widgets
# -----------------------------------------------------------------------

sub zoom_in {
  return if $zoom->ex <= $zoom_range[0];
  $zoom--;
  clear_map();
  draw_map();

  save_status();
}

sub zoom_out {
  return if $zoom->ex >= $zoom_range[1];
  $zoom++;
  clear_map();
  draw_map();

  save_status();
}

sub zoom_update {
  $zoom->update;
  clear_map();
  draw_map();

  save_status();
}

sub mag_in {
  return if $magnify->ex <= $magnify_range[0];
  $magnify++;
  clear_map();
  draw_map();

  save_status();
}

sub mag_out {
  return if $magnify->ex >= $magnify_range[1];
  $magnify--;
  clear_map( -keep_select => 1 );
  draw_map();

  save_status();
}


sub mag_update {
  $magnify->update;
  clear_map( -keep_select => 1 );
  draw_map();

  save_status();
}

sub toggle_cursor {
  $cursor_mode = $cursor_mode ? 0 : 1;
  update_cursor();

  save_status();
}

sub update_cursor {
  if ( $cursor_mode ) {
    update_scroll();
  } else {
    $canvas->coords( 'cursor_v', 0, 0, 0, 0 );
    $canvas->coords( 'cursor_h', 0, 0, 0, 0 );
  }

  save_status();
}

sub update_lnglat {
  my @offset = @_;
  @offset = ( 0, 0 ) unless @offset;
  my ( $lng, $lat ) = coords2lnglat( $map_coords[0] - $offset[0], $map_coords[1] - $offset[1] );

  $lnglat = sprintf( "%.6f,%.6f", $lng, $lat );
}

sub move_lnglat {
  my ( $lng, $lat ) = split( /,/, $lnglat );

  $lng = normalize_lnglat( $lng );
  $lat = normalize_lnglat( $lat );

  if ( !defined $lng || !defined $lat ) {
    # illegal input .. restore $lnglat here
    ( $lng, $lat ) = coords2lnglat( @map_coords );
    $lnglat = sprintf( "%.6f,%.6f", $lng, $lat );
    $canvas->Tk::focus;
    return;
  }
  $lnglat = "$lng,$lat";

  @map_coords = lnglat2coords( $lng, $lat );

  clear_map();
  draw_map();

  save_status();
}

sub normalize_lnglat {
  local( $_ ) = @_;

  return unless defined $_;

  if ( /^\d+$/ || /^\d*\.\d+$/ || /^\d+\.\d*$/ ) {
    return $_;
  } elsif ( /^[\d\.]*\.\d+\.[\d\.]*$/ ) {
    # degree.minute.second
    my @value = split( /\./ );
    my $r = 1;
    my $result = 0;
    foreach my $value ( @value ) {
      $result += $value * $r;
      $r /= 60;
    }
    return $result;
  } else {
    # illegal format
    return undef;
  }
}

sub toggle_border {
  $border_mode = $border_mode ? 0 : 1;
  update_border();

  save_status();
}

sub select_all {
  foreach my $map ( values %map ) {
    if ($map->{backup} && !$map->{selected} ) {
      $canvas->itemconfigure( $map->{border_id}, -outline => 'red' );
      $map->{selected} = 1;
    }
  }

  save_status();
}

sub update_border {
  if ( $border_mode ) {
    $canvas->raise( 'border', 'data' );
  } else {
    $canvas->lower( 'border', 'data' );
  }

  save_status();
}

sub move_map {
  my @diff = ( $_[1], $_[2] );

  # udpate mapCoords
  @map_coords = ( $map_coords[0] - $diff[0] * $zoom->p2 / $magnify->p2,
		  $map_coords[1] - $diff[1] * $zoom->p2 / $magnify->p2 );

  draw_map();

  save_status();

}

sub show_help {
  my $help_message =<<EOHelp;
pgmap ver $VERSION

I,Double click: Zoom in
O : Zoom out
Shift-I : Magnify
Shift-O : Shrink
Shift-Double click : Move to center

. : Toggle cursor
! : Toggle grid

Shift-D : Download maps
Shift-click : Select map
Shift-A : Select all maps

J : Input longtitude and latitude
A : Input address

Shift-Q : Quit program
Shift-R : Recover previous status
H,? : Show help

Cursor keys : Move map
Shift-cursor keys : Move map faster

EOHelp

  my $buttons = -d $basedir ? ['Dismiss'] : ['Initial map','Dismiss'];

  my $dialog = $top->DialogBox( -title => 'Help',
                           -buttons => $buttons, -default_button => 'Dismiss' );

  my $message = $dialog->add( 'Message', -text => $help_message );
  $message->pack();
  my $ret = $dialog->Show;

  if ( $ret =~ /initial/i ) {
    quickstart_download();
  }
}

sub address_entry {
  my $addr;

  my $dialog = $top->DialogBox( -title => 'Input address',
                           -buttons => ['OK'], -default_button => 'OK' );
  my $entry = $dialog->add( 'Entry', -textvariable => \$addr );
  $entry->pack();
  $dialog->Show;

  $status_message = 'Performing Address lookup';
  init_address_lookup() or return;
  $address_lookup_queue->enqueue( $addr );
}

# -----------------------------------------------------------------------
#  utilities
# -----------------------------------------------------------------------

sub clear_map {
  my ( %args ) = @_;

  return unless @canvas_center;

  $canvas->delete( 'data', 'border', 'poi' );

  # release resources assoicated with image
  my @caller = caller;
  foreach my $map ( values %map ) {
    if ( defined $map->{image} && !$map->{blank} ) {
      $map->{image}->delete;
      $debug_images--;
    }
  }

  if ( $args{-keep_select} ) {
    foreach my $xy ( keys %map ) {
      $map{$xy} = { selected => $map{$xy}->{selected} };
    }
  } else {
    %map = ();
  }
  %map_id = ();

  foreach my $data ( values %poi_id ) {
    $data->{cid} = undef;
  }
  %poi_id = ();

}

sub getImgURL {
  my ( $znum, $xnum, $ynum ) = @_;

  my $xmnum = 2 ** ( $zoom_range[1] - 1 - $znum );
  my $xlnum = $xnum >= 0 ? $xnum : $xnum + ceil( -$xnum / $xmnum ) * $xmnum;
  my $zstr = substr( "00" . $znum, -2 );
  my $xstr = substr( "0000000000" . ( $xlnum % $xmnum ), -10 );
  my $ystr = substr( "0000000000" . $ynum, -10 );
  my $imgURL = "$basedir/$zstr";
  foreach my $i ( 0 .. 9 ) {
    $imgURL .= "/" . substr( $xstr, $i, 1 ) . substr( $ystr, $i, 1 );
  }
  $imgURL . ".png";
}

sub lnglat2coords {
  my ( $lng, $lat ) = @_;
#  print "lnglat = @_\n";
  my @coords;

  my $zoom_imgsize = $img_size * 2 ** ( $zoom_range[1] - 1 );

  # normlization for lng
  $lng -= 360 while $lng >= 180;
  $lng += 360 while $lng < -180;
  $coords[0] = ( $lng + 180 ) / 360 * $zoom_imgsize;

  # normlization for lat
  $lat -= 180 while $lat >= 90;
  $lat += 180 while $lat < -90;
  $coords[1] = ( 0.5 - log( tan( ( $lat + 90 ) / 180 * pi() / 2 ) ) / ( 2 * pi() ) ) * $zoom_imgsize;

  ( floor( $coords[0] ), floor( $coords[1] ) );
}


sub coords2lnglat {
  my ( @coords ) = @_;

  my $zoom_imgsize = $img_size * 2 ** ( $zoom_range[1] - 1 );

  ( $coords[0] / $zoom_imgsize * 360 - 180,
    atan( exp( ( 0.5 - $coords[1] / $zoom_imgsize ) * 2 * pi() ) ) / pi() * 180 * 2 - 90 );
}

sub save_status {
  return unless $save_status_mode;

  local $" = ',';

  my $status = $running;
  my $selected = '';

  foreach my $xy ( sort sortxy keys %map ) {
    if ( $map{$xy}->{selected} ) {
      my ( $x, $y ) = split( /$;/, $xy );
      $selected .= "$x,$y;"
    }
  }
  chop $selected;

  open my $fh, '>', 'pgmap-save' or return;

print $fh <<EOC;
# this file is automatically generated

\@map_coords=(@map_coords);
\$cursor_mode=$cursor_mode;
\$border_mode=$border_mode;
\$zoom->set(@{[$zoom->ex]});
\$magnify->set(@{[$magnify->ex]});
\$running=$running;
\$geometry="@{[$top->geometry]}";
\$selected="$selected";
EOC

  close $fh;
}


sub load_prev_status {
  open my $fh, '<', 'pgmap-save' or return;
  @prev_status = <$fh>;
  close $fh;
}

sub apply_prev_status {
  return unless @prev_status;

  # load and recover

  my $geometry;
  my $selected;
  my $status;

  foreach ( @prev_status ) {
    chomp;
    s/\x0d$//;

    #remove 'running' entry here?
    eval( $_ );
  }

  clear_map();
  if ( defined $selected ) {
    foreach my $x_y ( split( /;/, $selected ) ) {
      my ( $x, $y ) = split( /,/, $x_y );
      $map{$x, $y}->{selected} = 1;
    }
  }

  if ( $geometry =~ /(\d+)x(\d+)/ ) {
    my ( $width, $height ) = ( $1, $2 );
    my $current_geometry = $top->geometry;
    $current_geometry =~ /(\d+)x(\d+)/;
    if ( $width != $1 || $height != $2 ) {
      $top->geometry( "${width}x${height}" );
      # automatically redraw is performed
      return;
    }
  }
  draw_map();
}

sub load_config {
  open my $fh, '<', 'pgmap.cfg' or return;
  my $poi_id = 0;
  @poi = ( { show => 1, name => 'default', file => undef, zoom => $zoom_range[1], icon => 'icon-red.png', id => $poi_id++ } );

  while ( <$fh> ) {
    chomp;
    s/\x0d$//;
    s/\#.*$//g;

    # poi entry here
    if ( /^\s*poi:\s*(.+)$/ ) {
      my @buf = split( /,/, $1 );
      push( @poi, { show => $buf[0], zoom => $buf[1], name => $buf[2], file => $buf[3], 
		    icon => $buf[4], desc=> $buf[5], id => $poi_id++ } );
      next;
    }

    my @config = split( /\s*=\s*/, $_ );
    if ( scalar( @config ) != 2 ) {
      next;
    }

    $config[0] =~ s/^\s+//;
    $config[1] =~ s/\s+$//;

    if ( $config[1] =~ /^[\s\d]+$/ ) {
      # only digit
#      print "config: \$$config[0]=$config[1]\n";
      eval( "\$$config[0]=$config[1]" );
    } else {
#      print "config: \$$config[0]=\"$config[1]\"\n";
      eval( "\$$config[0]=\"$config[1]\"" );
    }
  }

  close $fh;
}

sub load_poi {
  my @colors = qw/yellow red blue purple skyblue orange green palegreen pink brown gray/;

  $poi_index = 1;
  my $csv = Text::CSV_XS->new ({ binary => 1 }) or
    die "Cannot use CSV: ".Text::CSV->error_diag ();

  foreach my $poi ( @poi ) {
    next unless defined $poi->{file};
    # error dialog instead of 'die'
    open my $fh, "<:encoding(shiftjis)", $poi->{file} or die "$poi->{file}: $!";

    while ( my $row = $csv->getline( $fh ) ) {
      #    print "$row->[0],$row->[1]\n";
      next if !$row->[0] || !$row->[1];
      my $coords = [lnglat2coords( $row->[0], $row->[1] )];
      push( @{$poi->{data}}, { lng => $row->[0], lat => $row->[1], name => $row->[2],
			       comment => $row->[3], coords => $coords, plid => $poi->{id},
			       color => $colors[int(rand() * scalar( @colors ))],
			     index => $poi_index++ } );
    }
    close $fh;
  }
}

sub set_mark {
  my $poi = $poi[0]; # default poi

  my ( $lng, $lat ) = split( /,/, $lnglat );
  push( @{$poi->{data}}, { lng => $lng, lat => $lat, name => undef, comment => undef, 
			  coords => [@map_coords], icon => undef, plid => 0, index => $poi_index++ } );
  draw_map();
}

sub show_poilist_win {
  return if defined $poilist_win;

  # create window
  $poilist_win = $top->Toplevel( -title => 'POI list' );
  $poilist_win->protocol( 'WM_DELETE_WINDOW', \&close_poilist_win );

  # create widget
  my @header = qw/+ ID Icon Title Zoom Description/;
  my $hlist = $poilist_win->Scrolled( 'HList', -columns => scalar( @header ), -header => 1,
				   -command => \&poilist_select,
				   -width => 60, -height => 30 )
    ->pack( -expand => 1, -fill => 'both' );
  foreach my $i ( 0 .. $#header ) {
    $hlist->headerCreate( $i, -text => $header[$i] );
  }

  # set data
  foreach my $poi ( @poi ) {
    my $cb = $hlist->Checkbutton;
    $cb->configure( -variable => \$poi->{show}, -command => [\&toggle_poilist, $poi->{id}]);
    $hlist->add($poi->{id});
    $hlist->itemCreate($poi->{id}, 0, -itemtype => 'window', -widget => $cb );
    $hlist->itemCreate($poi->{id}, 1, -text => $poi->{id} );
    $hlist->itemCreate($poi->{id}, 2, -itemtype => 'image', -image => get_icon( $poi->{icon} ) );
    $hlist->itemCreate($poi->{id}, 3, -text => $poi->{name} );
    $hlist->itemCreate($poi->{id}, 4, -text => $poi->{zoom} );
    $hlist->itemCreate($poi->{id}, 5, -text => $poi->{desc} );
  }
}

sub toggle_poilist {
  my $poi = $poi[$_[0]];

  if ( $poi->{show} ) {
#    print "show poilist '@_'\n";
    # only update when turned on
    draw_map();
    return;
  }
#  print "delete poilist '@_'\n";

  foreach my $id ( keys %poi_id ) {
    next unless $poi_id{$id}->{plid} == $poi->{id};

    $canvas->delete( $id );
    $poi_id{$id}->{cid} = undef;
    delete $poi_id{$id};
  }

  update_poidata();
}

sub show_poidata_win {
  return if defined $poidata_win;
  $poidata_win = $top->Toplevel( -title => 'POI data' );
  $poidata_win->protocol( 'WM_DELETE_WINDOW', \&close_poidata_win );

  my @header = qw/Location ID Icon Name Description/;

  $poidata_hlist = $poidata_win->Scrolled( 'HList', -columns => scalar( @header ), -header => 1, 
				  -browsecmd => \&poidata_browse, -command => \&poidata_select,
				  -width => 60, -height => 30, -selectmode => 'browse' )
    ->pack( -expand => 1, -fill => 'both' );

  foreach my $i ( 0 .. $#header ) {
    $poidata_hlist->headerCreate($i, -text => $header[$i]);
  }

  $poi_select = -1;
  update_poidata();
}

sub update_poidata {
  return unless defined $poidata_hlist;

  $poidata_hlist->delete( 'all' );

  my $angle_num = 24;
  foreach my $data ( values %poi_id ) {
    $data->{dist} = sprintf( "%.1f", sqrt( ( $map_coords[0] - $data->{coords}->[0] ) ** 2 +
			    ( $map_coords[1] - $data->{coords}->[1] ) ** 2 ) / 4 / 1000 );
    $data->{angle} = floor( atan2( $data->{coords}->[1] - $map_coords[1], $data->{coords}->[0] - $map_coords[0] ) / ( 2 * pi() ) * $angle_num - 0.5 ) % $angle_num;
#    print "coords = @map_coords, data = @{$data->{coords}}, diff = @diff, angle = $data->{angle}, d$data->{name}\n";
  }

  foreach my $poi_id ( sort { $poi_id{$a}->{dist} <=> $poi_id{$b}->{dist} } keys %poi_id ) {
    $poidata_hlist->add($poi_id);
    $poidata_hlist->itemCreate($poi_id, 0, -itemtype => 'imagetext', 
			       -image => get_icon( "dir-$poi_id{$poi_id}->{angle}.png" ),
			       -text => "$poi_id{$poi_id}->{dist}km" );
    $poidata_hlist->itemCreate($poi_id, 1, -text => $poi_id{$poi_id}->{plid} );
    $poidata_hlist->itemCreate($poi_id, 2, -itemtype => 'imagetext',
			       -image => get_icon( $poi[$poi_id{$poi_id}->{plid}]->{icon} ) );
    $poidata_hlist->itemCreate($poi_id, 3, -text => $poi_id{$poi_id}->{name} );
    $poidata_hlist->itemCreate($poi_id, 4, -text => $poi_id{$poi_id}->{comment} );

    if ( $poi_id{$poi_id}->{index} == $poi_select ) {
#      print "found poi select $poi_select\n";
      $poidata_hlist->anchorSet( $poi_id );
    }
  }
}

sub poidata_browse {
  if ( @_ == 1 ) {
#    print "id $_[0] clicked (coords = @{$poi_id{$_[0]}->{coords}})\n";
    @poi_target = @{$poi_id{$_[0]}->{coords}};
    $poi_select = $poi_id{$_[0]}->{index};
#    print "selected id = $poi_select\n";
    update_scroll();
    $canvas->raise( 'poitarget', 'partition_data' );
    $canvas->lower( 'poitarget', 'partition_poi' );
  }
}

sub poidata_select {
  # edit poidata
}

sub poilist_select {
  # edit poilist
}

sub close_poilist_win {
  $poilist_win->destroy;

  $poilist_win = undef;
}

sub close_poidata_win {
  $poidata_win->destroy;

  $poidata_hlist = undef;
  $poidata_win = undef;

  @poi_target = ( 0, 0 );
  $poi_select = -1;
  $canvas->coords( 'poitarget', 0, 0, 0, 0 );
}

sub sortxy {
  my ( $ax, $ay ) = split( /$;/, $a );
  my ( $bx, $by ) = split( /$;/, $b );
  $ay <=> $by || $ax <=> $bx;
}

sub debug_func {
}

# ----------------------------------------------------------------------------

{
  package Log2;

  use overload
    '++' => \&increment,
      '--' => \&decrement;

  sub new {
    my ( $class, $value ) = @_;
    my $this = {
		value => 2 ** $value,
		exp => $value,
	       };
    bless $this, ref $class || $class;
  }

  sub set {
    my ( $this, $value ) = @_;

    $this->{exp} = $value;
    $this->{value} = 2 ** $value;
  }

  sub ex {
    my ( $this ) = @_;

    $this->{exp};
  }

  sub p2 {
    my ( $this ) = @_;

    $this->{value};
  }

  sub increment {
    my ( $this ) = @_;

    $this->set( $this->{exp} + 1 );
  }

  sub decrement {
    my ( $this ) = @_;

    $this->set( $this->{exp} - 1 );
  }

  sub update {
    my ( $this ) = @_;

    $this->{value} = 2 ** $this->{exp};
  }
}
